@@ -4,6 +4,7 @@
 
 use num_cpus;
 use std::os::raw::c_void;
+use std::sync::Arc;
 use std::thread;
 
 mod bindings;
@@ -12,24 +13,45 @@ use bindings::*;
 #[derive(Debug, Clone, Copy, PartialEq)]
 pub struct Output([u8; RANDOMX_HASH_SIZE as usize]);
 
+impl AsRef<[u8]> for Output {
+    fn as_ref(&self) -> &[u8] {
+        &self.0
+    }
+}
+
+impl Output {
+    pub fn leading_zeros(&self) -> u32 {
+        let mut zeros = 0;
+        for limb in self.0.iter().rev() {
+            let limb_zeros = limb.leading_zeros();
+            zeros += limb_zeros;
+            if limb_zeros != 8 {
+                break;
+            }
+        }
+        zeros
+    }
+}
+
 #[derive(Clone)]
 struct Sendable<T>(*mut T);
 unsafe impl<T> Send for Sendable<T> {}
 
-pub struct Hasher {
+pub struct Context {
+    flags: randomx_flags,
     fast: bool,
     cache: *mut randomx_cache,
-    vm: *mut randomx_vm,
     dataset: *mut randomx_dataset,
 }
-impl Hasher {
+
+impl Context {
     pub fn new(key: &[u8], fast: bool) -> Self {
         unsafe {
             let mut flags = randomx_get_flags();
             let mut cache = randomx_alloc_cache(flags);
             randomx_init_cache(cache, key.as_ptr() as *const c_void, key.len() as u64);
             let mut dataset = std::ptr::null_mut();
-            let vm = if fast {
+            if fast {
                 flags = flags | randomx_flags_RANDOMX_FLAG_FULL_MEM;
                 dataset = randomx_alloc_dataset(flags);
                 let num_threads = num_cpus::get();
@@ -56,19 +78,46 @@ impl Hasher {
 
                 randomx_release_cache(cache);
                 cache = std::ptr::null_mut();
-                randomx_create_vm(flags, std::ptr::null_mut(), dataset)
-            } else {
-                randomx_create_vm(flags, cache, std::ptr::null_mut())
-            };
+            }
 
-            Hasher {
+            Self {
+                flags,
                 fast,
                 cache,
-                vm,
                 dataset,
             }
         }
     }
+}
+
+impl Drop for Context {
+    fn drop(&mut self) {
+        unsafe {
+            if self.fast {
+                randomx_release_dataset(self.dataset);
+            } else {
+                randomx_release_cache(self.cache);
+            }
+        }
+    }
+}
+
+pub struct Hasher {
+    _context: Arc<Context>,
+    vm: *mut randomx_vm,
+}
+
+unsafe impl Send for Hasher {}
+
+impl Hasher {
+    pub fn new(context: Arc<Context>) -> Self {
+        unsafe {
+            Hasher {
+                _context: Arc::clone(&context),
+                vm: randomx_create_vm(context.flags, context.cache, context.dataset),
+            }
+        }
+    }
 
     pub fn hash(&self, inp: &[u8]) -> Output {
         let mut hash = [0u8; RANDOMX_HASH_SIZE as usize];
@@ -88,11 +137,6 @@ impl Drop for Hasher {
     fn drop(&mut self) {
         unsafe {
             randomx_destroy_vm(self.vm);
-            if self.fast {
-                randomx_release_dataset(self.dataset);
-            } else {
-                randomx_release_cache(self.cache);
-            }
         }
     }
 }
@@ -110,13 +154,13 @@ mod tests {
 
     #[test]
     fn test_slow_hasher() {
-        let slow = Hasher::new(KEY, false);
+        let slow = Hasher::new(Arc::new(Context::new(KEY, false)));
         assert_eq!(slow.hash(INPUT), EXPECTED);
     }
 
     #[test]
     fn test_fast_hasher() {
-        let fast = Hasher::new(KEY, true);
+        let fast = Hasher::new(Arc::new(Context::new(KEY, true)));
         assert_eq!(fast.hash(INPUT), EXPECTED);
     }
 }
